#include "debugger.h"
   
#define P(i) (stack[ind[i+1]].buf)
#define D(i) (stack[ind[i+1]].desc)
char *toLowerCase(char *exfuncName);
void *makeFuncCall(void *ind);
char (*fptr)();
extern int  nextModule;

void free_vars(char *usedVars[],int nVars)
{
  int k;
   /* free all the vars */
  for(k=0;k<nVars;k++)
    free(usedVars[k]);
}
  

int callFuncDynamic(char *curHpfFunc, char *userHpfStmts,ERRBUF *errbuf)
{
    int pid,status,numOfusedVars;
    char newFuncName[MAXVARLEN],cmnd[MAXVARLEN];
    char *usedVars[64];
    void *handle;
    char exfuncName[MAXVARLEN];
    pthread_t	tid;
    int ind[32],nparams,n,retval=0;
    sigset_t new;
    char newLibName[MAXVARLEN],newModuleName[MAXVARLEN];
    char *userDir,interpSoFile[MAXVARLEN];
    
    userDir = getenv("USERDIR");
    if (userDir == NULL)
      errexit("ME:%d,userdir NULL !!!. Exiting \n",ME);

    sprintf(interpSoFile,"%sfoos_%d.f",userDir,ME);
    retval = GetInstrumentedCode(nextModule,curHpfFunc,usedVars,
			 &numOfusedVars,userHpfStmts);   
    if (retval != 0){
      free_vars(usedVars,numOfusedVars);
      /* There is syntax error, leave it as it is and report error to user */
      sprintf(cmnd,"pghpf -Mfree -c %s > %serrout 2>&1",interpSoFile,userDir);
      /* pghpf -Mfree -c interpSoFile > errout 2>&1  */
      system(cmnd); 
      errbuf->type = COMPILE_ERROR;
      errbuf->status = retval;
      return retval;
    }
    
    sprintf(newModuleName,"newmodule_%d_%d",nextModule,ME); 
    sprintf(newFuncName,"newfunction_%d",nextModule);
    sprintf(newLibName,"%slibfoos_%d_%d.so",userDir,nextModule,ME);

    /* Build shared object from the on-the-fly generated function */
    sprintf(cmnd,"pghpf -Mfree -G %s -o %s > %serrout 2>&1",interpSoFile,
               newLibName,userDir);
    if (DEBUG)
      fprintf(fpout,"I execute this: %s retval = %d\n",cmnd,retval);
    retval = system(cmnd);
    if (DEBUG)
      fprintf(fpout,"I succesfully produced .so file for current func. \n");

    if (retval != 0){
      free_vars(usedVars,numOfusedVars);
      errbuf->type = COMPILE_ERROR;
      errbuf->status = retval;
      return retval;
    }  

    /* Fortran produces new function name in symbol table as 
       module_name+new_function_name */   
    /* Try for only subroutine not module */
    sprintf(exfuncName,"%s_%s_",newModuleName,newFuncName);
    
    if ((handle = dlopen(newLibName,RTLD_LAZY|RTLD_GLOBAL)) == NULL)
       fatal("I can not open library:%s:%s\n"," ",dlerror());
   
    if ((fptr = (char (*)())dlsym(handle,toLowerCase(exfuncName))) == NULL)
       fatal("I can not find function in .so:%s:%s\n",
	     toLowerCase(exfuncName),dlerror());

    if (DEBUG)
        fprintf(fpout,"After I opened shared object and found variable \n");
    
    nparams = calcIndices(newFuncName,usedVars, numOfusedVars,&ind[1]);
    ind[0]  = nparams;

    sigfillset(&new);
    pthread_sigmask(SIG_BLOCK, &new, NULL);
    if ( n = pthread_create( &tid, NULL, makeFuncCall, (void *)ind ) ) {
	fprintf( stderr, "pthread_create: %s\n", strerror( n ) );
	exit( 1 );
    }

    if ( n = pthread_join( tid, NULL ) ) {
	fprintf( stderr, "pthread_join: %s\n", strerror( n ) );
	exit( 1 );
    } 
    /* dlclose(handle); */
    if (DEBUG)
        fprintf(fpout,"After I executed function call \n");

    /* If I successfully executed previous interpretetation,
       increase module number */
    if (funcCallStatus == 0) 
      nextModule++;

    free_vars(usedVars,numOfusedVars);
    errbuf->type = RUNTIME_ERROR;  
    errbuf->status = funcCallStatus ;
    if (errbuf->status != 0)
      return (errbuf->status);
    else
      return 0;      
}
   
char *toLowerCase(char *s)
{
  int i;

  for(i=0;i<strlen(s);i++){
     if ((s[i]>=65)&&(s[i]<=90))
       s[i] += 32;
  }
  return s;
}


void  interrupt(int sig)
{
   funcCallStatus = sig;
   pthread_exit( 0 ); 
   return;
}

void *makeFuncCall(void *indx)
{
  int nparams,*ind = (int *)indx;
  struct sigaction act;  
  sigset_t new;  
  int n;

  funcCallStatus = 0;
  set_signal_handler(interrupt);
  nparams = ind[0];

  switch(nparams)
    {
    case 0: (*fptr)();
      break;
    case 1:  (*fptr)(P(0),D(0)); 
      break;
    case 2: (*fptr)(P(0),P(1),D(0),D(1)); 
     break;
    case 3:(*fptr)(P(0),P(1),P(2),D(0),D(1),D(2));
     break;
    case 4:(*fptr)(P(0),P(1),P(2),P(3),D(0),D(1),D(2),D(3));
     break;
    case 5:(*fptr)(P(0),P(1),P(2),P(3),P(4),D(0),D(1),D(2),D(3),D(4));
     break;
    case 6:(*fptr)(P(0),P(1),P(2),P(3),P(4),P(5),
           D(0),D(1),D(2),D(3),D(4),D(5));
    break;
    case 7:(*fptr)(P(0),P(1),P(2),P(3),P(4),P(5),P(6),
           D(0),D(1),D(2),D(3),D(4),D(5),D(6));
    break;
    case 8:(*fptr)(P(0),P(1),P(2),P(3),P(4),P(5),P(6),P(7),
           D(0),D(1),D(2),D(3),D(4),D(5),D(6),D(7));
    break;
    case 9:(*fptr)(P(0),P(1),P(2),P(3),P(4),P(5),P(6),P(7),P(8),
           D(0),D(1),D(2),D(3),D(4),D(5),D(6),D(7),D(8));
    break;
    case 10:(*fptr)(P(0),P(1),P(2),P(3),P(4),P(5),P(6),P(7),P(8),P(9),
           D(0),D(1),D(2),D(3),D(4),D(5),D(6),D(7),D(8),D(9));
    break;
    case 11:
       (*fptr)(P(0),P(1),P(2),P(3),P(4),P(5),P(6),P(7),P(8),P(9),P(10),
           D(0),D(1),D(2),D(3),D(4),D(5),D(6),D(7),D(8),D(9),D(10));
    break;
    case 12:
    (*fptr)(P(0),P(1),P(2),P(3),P(4),P(5),P(6),P(7),P(8),P(9),P(10),P(11),
       D(0),D(1),D(2),D(3),D(4),D(5),D(6),D(7),D(8),D(9),D(10),D(11));
    break;
 
    case 13:
    (*fptr)(P(0),P(1),P(2),P(3),P(4),P(5),P(6),P(7),P(8),P(9),P(10),P(11),
        P(12),D(0),D(1),D(2),D(3),D(4),D(5),D(6),D(7),D(8),D(9),D(10),D(11),
	    D(12));
 
    break;
    
    }  
    pthread_exit( 0 ); 
}
 
int  calcIndices(char *curHpfFunc,char *usedVars[],int numOfusedVars,int *ind)
{
   int i,r,cnt;

   for (i=0,cnt=0;i<numOfusedVars;i++)
   {
     if (((r = find_var(curFuncName,usedVars[i])) < 0)&&
	 ((r = find_var(curHpfFunc,usedVars[i])) < 0)) {
        if (DEBUG)
          fprintf(fpout,"I couldn't find the var : %s in function :%s\n",
		usedVars[i], curHpfFunc);
     }
     else{
        ind[cnt++] = r;
        if (DEBUG)
          fprintf(fpout,"PARAM : %s\n",usedVars[i]);
     }
   }
   if (DEBUG)
      fprintf(fpout,"After calcIndices nvars=%d\n",cnt);
   return cnt;
}
    
void fatal(char *s1, char *s2,char *s3)
{
  
  printf(s1,s2,s3);
  stop_program(1,2);
  
  printf("ME : %d exiting \n",ME);
  exit(3);
}

