#include "debugger.h"
#define INADDR_NONE 0xffffffff

/* client_type specifies that it is Applet,
    one of parallel nodes or another manager */
static fd_set	rset, allset;
static client[QLEN],client_type[QLEN];
static node_sockets[QLEN];
static int maxi,maxfd;
static int SageServerSock;
static int applicationType;
static num_of_pnodes = 0;

FILE *fpout = stdout;

int process_stop_ack(char *buf,int listenfd);

/********************************/
/* broadcast the command to all the registered clients 
if cli_type is APPLET,broadcast command to the registered applets 
if cli_type is PARNODE,broadcast command to the registered pnodes.
/********************************/
void bcast_command(int command,int listenfd,int cli_type)
{
   int i,sockfd,buflen,isStop;
   char *buf;
      
   recv_int(listenfd,&buflen); /* receive buffer length */
   buf = (char *)malloc(buflen);
   recv_raw_data(listenfd, buf, buflen);

   for (i = 0; i <= maxi; i++){
     if ((sockfd = client[i]) < 0)
	continue;
     if (client_type[i] == cli_type){
     /*Dont broadcast applet command to actual sender of this applet command*/
        if (listenfd != client[i]){
	  if (command  == JSWINDOW)
	    send_int(sockfd,JSWINDOW);
	  send_raw_data(sockfd,buf,buflen);
	}
	
     }
   }
   free(buf);
}

int close_parnode_sock(int ind)
{
     if (client_type[ind] != PARNODE)
       return;
          
     num_of_pnodes--;
     if (num_of_pnodes == 0){ /* all of parallel nodes exited */
       printf("MANAGER is exiting !!\n");
       close(SageServerSock);/* Close connection to sage server */
       exit(0);
     }
}


void start_parallel_application_and_sageServer(int sockfd)
{
  typedef char BUFTYPE[256];
  
  BUFTYPE script,HPFApplication,ManagerServerPort,SageServerHost;
  BUFTYPE SageServerPort,SciVisHost,SciVisPort,numOfProcs,appType;
  BUFTYPE ManagerHostName;
  char *command;
  int i;
    
  if (gethostname(ManagerHostName,256) < 0)
    errexit("I coult not get manager host name \n");
     
  recv_String(sockfd,script);
  recv_String(sockfd,HPFApplication);
  recv_String(sockfd,ManagerServerPort);
  recv_String(sockfd,SageServerHost);
  recv_String(sockfd,SageServerPort);
  recv_String(sockfd,SciVisHost);
  recv_String(sockfd,SciVisPort);
  recv_String(sockfd,numOfProcs);
  recv_String(sockfd,appType);
  applicationType = appType[0]-'0';
  
  command = (char *)malloc(5048);
  /* First Start Sage Server  */
  sprintf(command,"startSageServer.pl %s %s ",SageServerPort,SageServerHost);
  if (fork() == 0) {/* child */
    system(command); 
    exit(0);
  }

  /* Second Connect to Sage Server */
  SageServerSock=connectsock(SageServerHost,SageServerPort,"tcp");
  printf("Manager connected to sage server : %d\n",SageServerSock);

  /***************************/
  /* save Sage Server socket */
  for (i = 0; i < QLEN; i++)
    if (client[i] < 0){
      	client[i] = SageServerSock;/* save descriptor */
        break;
    }
  FD_SET(SageServerSock, &allset);/* add new descriptor to set */
  if (SageServerSock > maxfd)
     maxfd = SageServerSock;/* for select */
  if (i > maxi)
     maxi = i;/* max index in client[] array */
  /****************************/

  /* Third,Start Parallel application */
  sprintf(command,"%s %s %s %s %s %s %s %s %s %s ",script,HPFApplication,
	  SageServerPort,SageServerHost,ManagerServerPort,ManagerHostName,
	  appType,SciVisPort,SciVisHost, numOfProcs);
  if (fork() == 0) {/* child */
    system(command);
    exit(0);
  }
  printf("after starting hpf application \n");
  free(command);
}


void forward_command(int insock,int outsock)
{
  int  buflen;
  char *buf;
  
  printf("forward fron %d to %d by Manager \n",insock,outsock);
  recv_int(insock,&buflen); /* receive buffer length */
  buf = (char *)malloc(buflen);
  recv_raw_data(insock, buf, buflen);
  send_raw_data(outsock,buf,buflen);    
  printf("Manager forwarded the command to Sage server insock-outsock:%d-%d\n",
	 insock,outsock);
  free(buf);
  
}




int process_manager_requests(int cindex)
{
   int new_command,buflen,type,source_command,subcommand;
   char *buf=NULL,rval;
   int  listenfd = client[cindex],pid;
   
   printf("Enter process_manager_requests cindex=%d fd=%d\n",
	  cindex,listenfd);
   
   /* recv new command from clients */
   rval = recv_int(listenfd,&new_command); 
   printf("MANAGER:received command : %d retval:%d\n",new_command,rval);
   
   if (rval < 0)
     return rval;

   switch(new_command){
      case REGISTER_CLIENT:  /* register client as applet,pnode or manager */
	recv_int(listenfd,&type);
	client_type[cindex] = type;
	if (type == PARNODE){
	  recv_int(listenfd,&pid);
	  node_sockets[pid] = listenfd;
	  num_of_pnodes++;
	}
	printf("Client has been registered as %d \n",type);
	break;
  
      /*recv ack. from pnodes and broadcast it to registered applet clients*/
      case ACKCOMMAND: 
	bcast_command(ACKCOMMAND,listenfd,APPLET);
	break;

      /*recv commands(reply) from applets(sageServer)
            and broadcast it to parallel nodes*/
      case APPCOMMAND:
	bcast_command(APPCOMMAND,listenfd,PARNODE);
	break;

      /* Receive the commands the the changes made only at client side 
	 from any client and broadcat it for other clinets to update their
	 applet */
      case JSWINDOW:
	bcast_command(JSWINDOW,listenfd,APPLET);
	break;
      
      /* start parallel application and sage server */
      case ADMIN:
	recv_int(listenfd,&subcommand);
	switch(subcommand)
	{
	   case START :
	     start_parallel_application_and_sageServer(listenfd);
	     break;
	}
	break;
      
      /* Forward hpf front-end server(sage server) commands to sage server */
      case SAGECOMMAND:
        printf("Manager got Sage Command ..\n");
	forward_command(listenfd,SageServerSock);
	break;

      case SAGECOMMAND_ACK:
	printf("Manager got Reply from Sage server ..\n");
	recv_int(listenfd,&pid);
	forward_command(listenfd,node_sockets[pid]);
	break;
	
      default:
	printf("UNKNOWN Manager Command !! : %d\n",new_command );
   }
   printf("END of  process_manager_requests cindex=%d fd=%d\n",
	  cindex,listenfd);

   return rval;
}


int main(int argc, char **argv)
{   
        char    *manager_service;
	int	i,connfd, listenfd,sockfd;
	int	nready;
	int	n;
	char	line[REGBUFSIZE];
	int	clilen;
	struct sockaddr_in	cliaddr, servaddr;
	struct timeval tval={0,0};


        printf("begin of server loop at MANAGER \n");
	if (argc < 2)
	   errexit("usage: manager tcp_port\n");
	manager_service    = argv[1];

	/* Build tcp manager server */
	listenfd = passivesock(manager_service,"tcp",QLEN);
	maxfd = listenfd;		/* initialize */
	maxi = -1;			/* index into client[] array */
	for (i = 0; i < QLEN; i++){
	     client[i]      = -1; /* -1 indicates available entry */
	     client_type[i] = -1;
	}
	FD_ZERO(&allset);
	FD_SET(listenfd, &allset);
        
	for ( ; ; ) {
		rset = allset;		/* structure assignment */
		nready = select(maxfd+1, &rset, NULL, NULL, &tval);

		if (FD_ISSET(listenfd, &rset)) {/* new client connection */
		     clilen = sizeof(cliaddr);
		     connfd = accept(listenfd,
                                  (struct sockaddr *) &cliaddr, &clilen);
		     if (connfd  < 0)
                         errexit("accept: %s\n", sys_errlist[errno]);
    
		     for (i = 0; i < QLEN; i++)
			if (client[i] < 0) {
			   client[i] = connfd;/* save descriptor */
			   break;
			}
		     if (i == QLEN)
			errexit("too many clients");
                     printf("Manager connected to client with fd=%d\n",connfd);
		     
		     FD_SET(connfd, &allset);/* add new descriptor to set */
		     if (connfd > maxfd)
			maxfd = connfd;	/* for select */
		     if (i > maxi)
			maxi = i;/* max index in client[] array */

		     if (--nready <= 0)
			continue;/* no more readable descriptors */
		}
                /* check all clients for data */
		for (i = 0; i <= maxi; i++) 
                {
		     if ( (sockfd = client[i]) < 0)
			continue;
		     if (FD_ISSET(sockfd, &rset)) {
		        n = process_manager_requests(i);
			if (n < 0){
			    printf("Manager closed the connection-fd:%d\n",
				  sockfd);
			   /* connection closed by client */
			    close(sockfd);
			    FD_CLR(sockfd, &allset);
			    client[i] = -1;
			    close_parnode_sock(i);
                        }
			if (--nready <= 0)
			  break;	/* no more readable descriptors */
		     } /* if ( */
		}/* for(i=0 */
	}/*for (;;) */

	printf("End of server loop in MANAGER \n");
	
}/*main*/

/*------------------------------------------------------------------------
 * passivesock - allocate & bind a server socket using TCP or UDP
 *------------------------------------------------------------------------
 */

int passivesock(char *service, char *protocol, int qlen )
/* char    *service;  */     /* service associated with the desired port     */
/* char    *protocol;  */    /* name of protocol to use ("tcp" or "udp")     */
/* int     qlen;      */     /* maximum length of the server request queue   */
{
        struct servent  *pse;   /* pointer to service information entry */
        struct protoent *ppe;   /* pointer to protocol information entry*/
        struct sockaddr_in sin; /* an Internet endpoint address         */
        int     s, type;        /* socket descriptor and socket type    */
	int ret;
	u_short  portbase = 0;  

        bzero((char *)&sin, sizeof(sin));
        sin.sin_family = AF_INET;
        sin.sin_addr.s_addr = INADDR_ANY;

    /* Map service name to port number */
        if ( pse = getservbyname(service, protocol) )
                sin.sin_port = htons(ntohs((u_short)pse->s_port)
                        + portbase);
        else if ( (sin.sin_port = htons((u_short)atoi(service))) == 0 )
                errexit("can't get \"%s\" service entry\n", service);

    /* Map protocol name to protocol number */
        if ( (ppe = getprotobyname(protocol)) == 0)
                errexit("can't get \"%s\" protocol entry\n", protocol);

    /* Use protocol to choose a socket type */
        if (strcmp(protocol, "udp") == 0)
                type = SOCK_DGRAM;
        else
                type = SOCK_STREAM;

    /* Allocate a socket */
        s = socket(PF_INET, type, ppe->p_proto);
        if (s < 0)
                errexit("can't create socket: %s\n", sys_errlist[errno]);

    /* Bind the socket */
        if (bind(s, (struct sockaddr *)&sin, sizeof(sin)) < 0)
                errexit("can't bind to %s port: %s\n", service,
                        sys_errlist[errno]);
        if (type == SOCK_STREAM && listen(s, qlen) < 0)
                errexit("can't listen on %s port: %s\n", service,
                        sys_errlist[errno]);
	
        return s;
}


void send_int(int sockfd,int command)
{
  char cmnd[4];
  XDR xdrs;
 
  xdrmem_create(&xdrs,cmnd,sizeof(int),XDR_ENCODE); 
  xdr_int(&xdrs,&command);
  send_raw_data(sockfd,cmnd,sizeof(int));

  xdr_destroy(&xdrs);
  if (DEBUG)
     fprintf(fpout,"After send INTEGER : %d\n",command);
}
void recv_String (int sockfd,char *variableName)
{
   bool_t ret;

   char *buffer,*tbuf;
   int strLen,totalLen;
   XDR xdrs;   
   int i;

   tbuf = (char *)malloc(REGBUFSIZE);
   buffer = (char *)malloc(REGBUFSIZE);

   recv_raw_data(sockfd,buffer,sizeof(int));
   xdrmem_create(&xdrs,buffer,sizeof(int),XDR_DECODE); 
   ret = xdr_int(&xdrs,&strLen);
   xdr_destroy(&xdrs);
   totalLen = strLen;
   if ((strLen%4) != 0)
     totalLen = strLen+4-strLen%4;
  
   recv_raw_data(sockfd,&buffer[sizeof(int)],totalLen);
   xdrmem_create(&xdrs,buffer,REGBUFSIZE,XDR_DECODE);
   ret = xdr_string(&xdrs,&tbuf,totalLen);
   tbuf[strLen] = 0; /* Append terminating character  */ 
   strcpy(variableName,tbuf);

   xdr_destroy(&xdrs);
   free(tbuf);
   free(buffer);
   if (DEBUG) 
      fprintf(fpout,"I received string : %s strlen = %d totlen=%d\n",
         variableName,strLen,totalLen);
   
}
int recv_int(int sockfd,int *command)
{
  char cmnd[4];
  XDR xdrs;
  int raw;
   
  raw = recv_raw_data(sockfd,cmnd,sizeof(int));
  if (raw <= 0)
    return -1;
  
  xdrmem_create(&xdrs,cmnd,sizeof(int),XDR_DECODE); 
  xdr_int(&xdrs,command);
  xdr_destroy(&xdrs);
  if (DEBUG)
    fprintf(fpout,"I received integer : %d\n",*command);
  return 0;
  
}
void send_raw_data(int sockfd,char *buf,int len)
{
  char *ptr;
  int nleft,nwritten;
  sigset_t oset;
 
  /* Mask three signals */
  /* mask_three_signals(&oset); */

  ptr   = buf;
  nleft = len;
  while (nleft > 0) {
    /*nwritten = write(sockfd, ptr, nleft); */
    nwritten = send(sockfd, ptr, nleft,0);
    if (nwritten<=0)
      errexit("send_buffer: nwritten <= 0");
    nleft -= nwritten;
    ptr += nwritten;
  }

  /* Unmask three signals */
  /* unmask_three_signals(&oset); */
  
  if (DEBUG)
     fprintf(fpout,"I send raw data sent len =%d\n",len);
}

int recv_raw_data(fd, buf, nbytes)
register int    fd;
register char   *buf;
register int    nbytes;
{
    int nleft, nread;
    char *ptr = buf;
    sigset_t oset;
    
    /* Mask three signals */
    /* mask_three_signals(&oset); */

    nleft = nbytes;
    while (nleft > 0) {
        /* nread = read(fd, ptr, nleft); */
	nread = recv(fd, ptr, nleft,0);
        if (nread < 0)
            return(nread);      /* error, return < 0 */
        else if (nread == 0)
            break;          /* EOF */

        nleft -= nread;
        ptr   += nread;
    }
    if (DEBUG)
       fprintf(fpout,
	       "%d bytes read for var =%d original = %d left =%d\n",ptr-buf,
	       fd,nbytes,nleft);
    
    /* Unmask three signals */
    /* unmask_three_signals(&oset); */
    return(nbytes - nleft);     /* return >= 0 */

}
int extract_int(char *buf)
{
  char cmnd[4];
  XDR xdrs;
  int command;  

  memcpy(cmnd,buf,sizeof(int));
  xdrmem_create(&xdrs,cmnd,sizeof(int),XDR_DECODE); 
  xdr_int(&xdrs,&command);

  xdr_destroy(&xdrs);
  return command;
  
  
}
 int errexit(char *format, ...)
 
{
    va_list args;
 
    va_start(args,format);
    _doprnt(format, args, stderr);
    va_end(args);
    exit(1);
}


/*------------------------------------------------------------------------
 * connectsock - allocate & connect a socket using TCP or UDP
 *------------------------------------------------------------------------
 */
int
connectsock( host, service, protocol )
char    *host;      /* name of host to which connection is desired  */
char    *service;   /* service associated with the desired port */
char    *protocol;  /* name of protocol to use ("tcp" or "udp") */
{
    struct hostent  *phe;   /* pointer to host information entry    */
    struct servent  *pse;   /* pointer to service information entry */
    struct protoent *ppe;   /* pointer to protocol information entry*/
    struct sockaddr_in sin; /* an Internet endpoint address     */
    int s, type;    /* socket descriptor and socket type    */
    int n;
     
    bzero((char *)&sin, sizeof(sin));
    sin.sin_family = AF_INET;
 
    /* Map service name to port number */
    if ( pse = getservbyname(service, protocol) )
        sin.sin_port = pse->s_port;
    else if ( (sin.sin_port = htons((u_short)atoi(service))) == 0 )
        errexit("can't get \"%s\" service entry\n", service);
 
    /* Map host name to IP address, allowing for dotted decimal */
    if ( phe = gethostbyname(host) )
 
        bcopy(phe->h_addr, (char *)&sin.sin_addr, phe->h_length);
    else if ( (sin.sin_addr.s_addr = inet_addr(host)) == INADDR_NONE )
        errexit("can't get \"%s\" host entry\n", host);
 
    /* Map protocol name to protocol number */
    if ( (ppe = getprotobyname(protocol)) == 0)
        errexit("can't get \"%s\" protocol entry\n", protocol);
 
    /* Use protocol to choose a socket type */
    if (strcmp(protocol, "udp") == 0)
        type = SOCK_DGRAM;
    else
        type = SOCK_STREAM;
 
    /* Allocate a socket */
    s = socket(PF_INET, type, ppe->p_proto);
    if (s < 0)
        errexit("can't create socket: %s\n", sys_errlist[errno]);
 
    /* For now waits for connection */
    /*
      if ((strcmp(protocol, "tcp") == 0)
	while(connect(s, (struct sockaddr *)&sin, sizeof(sin)) < 0); 
    */
    
    /* Connect the socket */
    if (strcmp(protocol, "tcp") == 0){
       if (connect(s, (struct sockaddr *)&sin, sizeof(sin)) < 0)
	{
	  /* if server is not started, reconnect again */
	  if (errno == ECONNREFUSED){ /* connection refused */
	    close(s);
	    connectsock( host, service, protocol );
	  }
	  else
	    errexit("can't connect to %s.%s: %s errno=%d\n", host, service,
		    sys_errlist[errno],errno);
	}
       
     }
     return s;  
}   
